# coding: utf-8
"""
 Die Klasse Grafikzeichner -- Version 2.1:
   Eine Erweiterung der Stifte und Maeusebibliothek
   fuer die Python-Portierung von Ingo Linkweiler
   mit kleinen eigenen Modifikationen vom 24.09.2008.
   
   Mit dieser Erweiterung koennen Objekte des
   Vektorgrafikpaketes PyObjVG gezeichnet werden.
   
   AUTOMATIKMODUS:
   Eine Verbindung zwischen Grafikzeichner und (Vektor-)Grafiken
   ermoeglicht es, Aenderungen von Attributen bei den
   beteiligten Grafikobjekten direkt auf dem Bildschirm
   sichtbar zu machen (Eine Neuzeichnung der Grafik muss
   nicht manuell angestossen werden. Andere Objekte koennen
   jedoch nicht auf dem Bildschirm dargestellt werden, da vor
   jedem Neuzeichnen, der Bildschirm automatisch geloescht wird).
   
   Autor: Matthias Heming <matthias@familie-heming.de>
   
   Dieses Klasse wird unter der Creative Commons by-nc-sa
   veroeffentlich, naehere Informationen dazu unter  
   http://creativecommons.org/licenses/by-nc-sa/3.0/de/.
   
   Versionsgeschichte:
     Version 1.0 vom 24. September 2008
     Version 2.0 vom 15. Januar 2009
     - Ueberarbeitung zur Python-New-Style-Klasse
     - Anpassung an Verbindungskonzept zu PyObjVG Version 2.0
     Version 2.1 vom 22. Januar 2009
     - Es findet keine Objekterkennung mehr statt, statt dessen
       wird im Grafikzeichner eine abstrahierte Zeichenschnittstelle
       dargeboten, ueber die sich die Objekte sich selbst zeichnen.
       => Damit ist die Integration von extern hergestellten
        Klassen vereinfacht. Diese muessen nur eine einzige Methode
        "zeichne(...)" zur Verfuegung stellen und koennen dann
        auf die vom Grafikzeichner abstrahierte Schnittstelle zugreifen.
     - Automatikmodus wird jetzt ueberall klein geschrieben.
       z.B. in der Methode gibAutomatikmodusStatus.
       Die Methode "setzeAutomatikModus" heisst jetzt "setzeAutomatikmodusStatus".
"""

class Grafikzeichner(object):
    """
    Der Grafikzeichner kennt -- analog zu der Klasse Stift -- einen Bildschirm.
    Auf diesen Bildschirm kann ein Grafikzeichner Objekte des Types "Vektorgraphik"
    zeichnen.
    Ein Grafikzeichner kann ausserdem mit genau einem Grafikobjekt direkt
    verbunden werden. Wird der Automatik-Modus aktiviert, so wird bei
    Aenderungssignalen des Grafikobjektes eine Neuzeichnung ausgeloest.
    Fuer die Ausgabe innerhalb des PyObjVG-Objektkartentext kann ein Name
    vergeben werden.
    """
    def __init__(self, bildschirm=None,name="Grafikzeichner"):
        """
        Initialisierung des Grafikzeichners.
        Als Parameter muss der zugehoerende Bildschirm angegeben werden.
        """
        if bildschirm != None:
          self.__bildschirm=bildschirm
        else:
          raise Exception("Es ist kein Bildschirm verfuegbar.")
        
        self.__ZugeordneteGrafik = None
        # keine Vektorgrafik verbunden
        self.__automatikmodus = False
        # es werden keine Aenderungssignale verarbeitet
        self.__freigegeben = True
        self.name = name
    
    def gibFrei(self):
        """
        Loest eine eventuell vorhandene Verbindung zu einer Vektorgrafik,
        setzt die interne Bildschirmreferenz auf 'None' und deaktiviert
        internes Freigegeben-Flag, so dass nachfolgende Aufrufe
        eine Ausnahme erzeugen.
        """
        if self.__ZugeordneteGrafik is not None:
            self.objektTrennen(self.__ZugeordneteGrafik)
        self.__bildschirm = None
        self.__freigegeben = False
    
    def setzeAutomatikmodusStatus(self,status):
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.')
        if status:
            self.__automatikmodus = True
            if self.__ZugeordneteGrafik is not None:
                self.zeichneGrafik(self.__ZugeordneteGrafik)
        else:
            self.__automatikmodus = False
    
    def gibAutomatikmodusStatus(self):
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.')
        return self.__automatikmodus
    
    automatikmodus = property(fget=gibAutomatikmodusStatus,fset=setzeAutomatikmodusStatus,fdel=None,doc=\
            """
            Status-Flag des automatischen Zeichenmodus:
            True:  Aenderungssignale vom verbundenen Objekt fuehren zur
                   automatisierten Neuzeichnung der Vektorgrafik auf dem
                   Bildschirm. Dazu wird der Bildschirm vorher geloescht.
            False: Keine automatische Neuzeichnung.
            """)
    def objektVerbinden(self,neuesobjekt):
        """
        Es wird das uebergebene Objekt mit dem Zeichner verbunden.
        
        Falls die Methode nicht ueber das uebergebene Objekt
        aufgerufen wurde, wird auch aus Sicht des Objekts
        die Verbindung hergestellt.
        Eine Verbindungsfehler-Ausnahme wird in einem der folgenden Faelle erzeugt:
        - Ein Objekt ist bereits verbunden
          (moeglich, falls die Methode nicht ueber das Objekt aufgerufen wurde)
        - Ein Objekt ist bereits zu einer anderen Gruppe verbunden
          (moeglich, falls die Methode nicht ueber das Objekt aufgerufen wurde)
        - Objekt stellt keine Methoden zur Verbindungskontrolle
          (verbindeMitObjekt//gibVerbundenesObjekt) zur Verfuegung.
        """
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.')
        try:
            if self.__ZugeordneteGrafik is not None:
                raise Verbindungsfehler("Es ist bereits ein Objekt mit dem Grafikzeichner verbunden.")
            con = neuesobjekt.gibVerbundenesObjekt()
            if (con is None):
                neuesobjekt.verbindeMitObjekt(self)
                # Verbindung wird auf Objektseite initialisiert
            elif (con==self):
                if (con == self.__ZugeordneteGrafik):
                    raise Verbindungsfehler("Objekt ist bereits verbunden.")
                else:
                    self.__ZugeordneteGrafik=neuesobjekt
                    self.objektGeaendert(neuesobjekt)
                    return True
            else:
                raise Verbindungsfehler("Objekt ist anderweitig verbunden.")      
        except AttributeError:
            raise Verbindungsfehler("Objekt stellt keine Methode zur Verbindungskontrolle zur Verfuegung.")
    
    def objektTrennen(self,objekt):
        """
        Es wird das verbundene Objekt vom Grafikzeichner entfernt.
        Ein Verbindungsfehler wird in einem der folgenden Faelle erzeugt:
        - Das Objekt war nicht mit der Gruppe verbunden.
        - Das Objekt stellt keine Methoden zur Verbindungskontrolle
          (trenneVonObjekt//gibVerbundenesObjekt) zur Verfuegung.
        """
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.')
        if objekt != self.__ZugeordneteGrafik:
            raise Verbindungsfehler("Objekt war nicht mit dieser Vektorgrafik verbunden.")
        try:
            con = objekt.gibVerbundenesObjekt()
            if con is not None:
                # eine Verbindung zu einer anderen Gruppe ist durch vorherige
                # Abfrage ausgeschlossen
                objekt.trenneVonObjekt()
                # Trennung auf Objektseite initialisiert
            else:
                self.__ZugeordneteGrafik = None
                return True
        except AttributeError:
            self.__ZugeordneteGrafik = None
            raise Verbindungsfehler("Das Objekt stellt keine Methoden zur Verbindungskontrolle zur Verfuegung.")
    
    def objektGeaendert(self,obj):
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.') # wird automatisiert von verbundenen Vektorgrafiken aufgerufen
        if self.__ZugeordneteGrafik == obj:
            if self.automatikmodus:
                self.__bildschirm.loescheAlles()
                self.zeichneGrafik(obj)
    
    def gibVerbundenesObjekt(self):
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.')
        return self.__ZugeordneteGrafik
    
    
    # Abstrahierte Zeichenroutinen
    def zeichnePolygon(self,pktliste,lfarbe,lstaerke,ffarbe):
        """
        Zeichnet einen Polygon gegeben durch Punkte in pktliste.
        Dabei muessen die Koordinaten jedes Punktes p der Liste
        mit int(p[0]) bzw. int(p[1]) lesbar sein.
        Weitere Vorgaben:
           lfarbe,ffarbe: Ganzzahlige RGB-Farbwerte, die als 3-Tupel gelesen
               werden koennen: int(xfarbe[i]), None, falls keine Farbe
               gezeichnet werden soll.
           lstaerke: wird mit int(...) gelesen
        => Die Polygonumrandung wird nicht gezeichnet, falls
           lfarbe = None oder lstaerke <= 0
        => Das Polygon ist nicht gefuellt, falls ffarbe=None
        """
        pktliste = [(int(pkt[0]),int(pkt[1])) for pkt in pktliste]
        if lfarbe is not None:
            lfarbe = int(lfarbe[0]),int(lfarbe[1]),int(lfarbe[2])
        lstaerke = int(lstaerke)
        if ffarbe is not None:
            ffarbe = int(ffarbe[0]),int(ffarbe[1]),int(ffarbe[2])
        # Achtung: Fuellung zuerst zeichnen, damit die Linien
        # die Fuellung ueberdeckt
        if pktliste != [ ] and ffarbe is not None:
            self.__bildschirm.bild.drawpolygon(pktliste,ffarbe,0)
        if pktliste != [ ] and lstaerke > 0 and lfarbe is not None:
            self.__bildschirm.bild.drawpolygon(pktliste,lfarbe,lstaerke)
    
    def zeichnePolygonzug(self,pktliste,lfarbe,lstaerke):
        """
        Zeichnet einen Polygonzug durch die in pktliste angegebenen
        Punkte. Dabei muessen die Koordinaten jedes Punktes p der Liste
        mit int(p[0]) bzw. int(p[1]) lesbar sein.
        Weitere Vorgaben:
           lfarbe: Ganzzahlige RGB-Farbwerte, die als 3-Tupel gelesen
               werden koennen: int(xfarbe[i])
           lstaerke: wird mit int(...) gelesen
        => Der Polygonzug wird nicht gezeichnet, falls
           lfarbe = None oder lstaerke <= 0
        """
        pktliste = [(int(pkt[0]),int(pkt[1])) for pkt in pktliste]
        if lfarbe is not None:
            lfarbe = int(lfarbe[0]),int(lfarbe[1]),int(lfarbe[2])
        lstaerke = int(lstaerke)
        if pktliste != [ ] and lstaerke > 0 and lfarbe is not None:
            self.__bildschirm.bild.drawpolyline(pktliste,lfarbe,lstaerke)
    
    def zeichneText(self,txt,mpkt,sgrad,winkel,lfarbe,lstaerke=1,ffarbe=None):
        """
        Zeichnet einen Text txt mit dem Mittelpunkt mpkt und gegebener
        Schriftgroesse. Ausser der Textfarbe wird noch keine veraenderte
        Linienstaerke oder Fuellfarbe unterstuetzt, die Parameter sind
        fuer zukuenftige Versionen vorgesehen.
        Vorgaben:
           txt: wird mit str(...) gelesen
           mpkt: Koordinaten muessen mit int(mpkt[i]) extrahiert werden koennen
           sgrad: wird mit int(...) gelesen
           winkel: wird mit int(...) gelesen und im Gradmass erwartet
           lfarbe,ffarbe: Ganzzahlige RGB-Farbwerte, die als 3-Tupel gelesen
               werden koennen: int(xfarbe[i]), None, falls keine Farbe
               gezeichnet werden soll.
           lstaerke: wird mit int(...) gelesen
        => Der Text wird nicht gezeichnet, falls lfarbe = None,
           lstaerke <= 0 oder sgrad <= 0
        """
        # Parameterinterpretationen
        txt = str(txt)
        mpkt = mpkt[0],mpkt[1]
        sgrad = int(sgrad)
        winkel = int(winkel)
        if ffarbe is not None:
            ffarbe = int(ffarbe[0]),int(ffarbe[1]),int(ffarbe[2])
        if lfarbe is not None:
            lfarbe = int(lfarbe[0]),int(lfarbe[1]),int(lfarbe[2])
        lstaerke = int(lstaerke)
        # Nutzung des Bildschirmserver zum Zeichnen
        if grad > 0 and lfarbe is not None and lstaerke > 0:
            self.__bildschirm.bild.drawtext(txt,mpkt,lfarbe,"",grad,drehwinkel)
                # TODO: Realisierung gefuellter Texte fehlt noch,
                # m. E. sinnvoll den Bildserver zu erweitern, mal sehen...
    
    def zeichneEllipse(self,mpkt,radx,rady,winkel,lfarbe,lstaerke,ffarbe):
        """
        Zeichnet eine Ellipse um den Mittelpunkt mpkt mit uebergegebenen
        Radien, Linienfarbe, -staerke und Fuellfarbe.
        Vorgaben:
           mpkt: Koordinaten muessen mit int(mpkt[i]) extrahiert werden koennen
           radx, rady: werden mit int(...) gelesen
           winkel: wird mit int(...) gelesen und im Gradmass erwartet
           lfarbe,ffarbe: Ganzzahlige RGB-Farbwerte, die als 3-Tupel gelesen
               werden koennen: int(xfarbe[i]), None, falls keine Farbe
               gezeichnet werden soll.
           lstaerke: wird mit int(...) gelesen
        => Die Ellipsenlinie wird nicht gezeichnet, wenn
           lfarbe=None oder lstaerke<=0 ist
        => Die Ellipse ist nicht gefuellt, falls ffarbe=None
        """
        # Parameterinterpretationen
        mpkt = mpkt[0],mpkt[1]
        winkel = int(winkel)
        if ffarbe is not None:
            ffarbe = int(ffarbe[0]),int(ffarbe[1]),int(ffarbe[2])
        if lfarbe is not None:
            lfarbe = int(lfarbe[0]),int(lfarbe[1]),int(lfarbe[2])
        radx,rady = int(radx),int(rady)
        lstaerke = int(lstaerke)
        # Nutzung des Bildschirmserver zum Zeichnen
        if ffarbe is not None:
            self.__bildschirm.bild.drawellipse(mpkt,radx,rady,ffarbe,0,winkel)
        if lstaerke > 0 and lfarbe is not None:
            self.__bildschirm.bild.drawellipse(mpkt,radx,rady,lfarbe,lstaerke,winkel)
    
    def zeichneGrafik(self,grafik):
        """
        Ruft die Methode "zeichnen" des uebergebenen Objektes mit sich selbst als
        Parameter auf. Dadurch werden die oben definierten Zeichenmethoden
        fuer das Grafikobjekt verfuegbar, welches den Rest selbst erledigen muss.
        """
        if not self.__freigegeben:
            raise Exception('Grafikzeichner ist nicht mehr verfuegbar.')
        else:
            grafik.zeichnen(self)
    

    
    

    